/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this file,
 * You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Portions created by Devexperts are Copyright (C) 2007-2012 Devexperts.
 */

package com.devexperts.evolution.utils.configencoder;

import java.io.InputStream;
import java.io.IOException;

import java.lang.ref.Reference;
import java.lang.ref.WeakReference;
import java.beans.ExceptionListener;

/**
 * The <code>ConfigDecoder</code> class is used to read textual documents created using the <code>ConfigEncoder</code>
 * and is used just like the <code>XMLDecoder</code>.
 *
 * @author Anton V. Shakhmin
 * @see java.beans.XMLDecoder
 * @see ConfigEncoder
 */

public class ConfigDecoder {
    private InputStream in;
    private Object owner;
    private ExceptionListener exceptionListener;
    private ConfigParser handler;
    private Reference<ClassLoader> clref;
    private NameConverter nameConverter = new NothingNameConverter();
    private ValueResolver valueResolver = new SimpleValueResolver();
    private ClassResolver classResolver = new ClassResolver();

    /**
     * Creates a new input stream for reading archives created by the <code>ConfigEncoder</code> class.
     *
     * @param in The underlying stream.
     *
     * @see ConfigEncoder#ConfigEncoder(java.io.OutputStream)
     */
    public ConfigDecoder(InputStream in) {
        this(in, null);
    }

    /**
     * Creates a new input stream for reading archives created by the <code>ConfigEncoder</code> class.
     *
     * @param in    The underlying stream.
     * @param owner The owner of this stream.
     */
    public ConfigDecoder(InputStream in, Object owner) {
        this(in, owner, null);
    }

    /**
     * Creates a new input stream for reading archives created by the <code>ConfigEncoder</code> class.
     *
     * @param in                the underlying stream.
     * @param owner             the owner of this stream.
     * @param exceptionListener the exception handler for the stream; if <code>null</code> the default exception
     *                          listener will be used.
     */
    public ConfigDecoder(InputStream in, Object owner, ExceptionListener exceptionListener) {
        this(in, owner, exceptionListener, null);
    }

    /**
     * Creates a new input stream for reading archives created by the <code>ConfigEncoder</code> class.
     *
     * @param in                the underlying stream.  <code>null</code> may be passed without error, though the
     *                          resulting ConfigDecoder will be useless
     * @param owner             the owner of this stream.  <code>null</code> is a legal value
     * @param exceptionListener the exception handler for the stream, or <code>null</code> to use the default
     * @param cl                the class loader used for instantiating objects. <code>null</code> indicates that the
     *                          default class loader should be used
     *
     * @since 1.5
     */
    public ConfigDecoder(InputStream in, Object owner,
        ExceptionListener exceptionListener, ClassLoader cl)
    {
        this.in = in;
        setOwner(owner);
        setExceptionListener(exceptionListener);
        setClassLoader(cl);
    }

    /**
     * Set the class loader used to instantiate objects for this stream.
     *
     * @param cl a classloader to use; if null then the default class loader will be used
     */
    private void setClassLoader(ClassLoader cl) {
        if (cl != null) {
            this.clref = new WeakReference<ClassLoader>(cl);
        }
    }

    /**
     * Return the class loader used to instantiate objects. If the class loader has not been explicitly set then null is
     * returned.
     *
     * @return the class loader used to instantiate objects
     */
    private ClassLoader getClassLoader() {
        if (clref != null) {
            return clref.get();
        }
        return null;
    }

    /**
     * This method closes the input stream associated with this stream.
     */
    public void close() {
        if (in != null) {
            try {
                in.close();
            } catch (IOException e) {
                getExceptionListener().exceptionThrown(e);
            }
        }
    }

    /**
     * Sets the exception handler for this stream to <code>exceptionListener</code>. The exception handler is notified
     * when this stream catches recoverable exceptions.
     *
     * @param exceptionListener The exception handler for this stream; if <code>null</code> the default exception
     *                          listener will be used.
     *
     * @see #getExceptionListener
     */
    public void setExceptionListener(ExceptionListener exceptionListener) {
        this.exceptionListener = exceptionListener;
    }

    // REVIEW: CIT: Should be final? Can be null?
    static ExceptionListener defaultExceptionListener = new ExceptionListener() {
        public void exceptionThrown(Exception e) {
            System.err.println(e);
            // e.printStackTrace();
            System.err.println("Continuing ...");
        }
    };

    /**
     * Gets the exception handler for this stream.
     *
     * @return The exception handler for this stream. Will return the default exception listener if this has not
     *         explicitly been set.
     *
     * @see #setExceptionListener
     */
    public ExceptionListener getExceptionListener() {
        return (exceptionListener != null) ? exceptionListener : defaultExceptionListener;
    }

    /**
     * Reads the next object from the underlying input stream.
     *
     * @return the next object read
     *
     * @throws ArrayIndexOutOfBoundsException if the stream contains no objects (or no more objects)
     * @see ConfigEncoder#writeObject
     */
    public Object readObject() {
        if (in == null) {
            return null;
        }
        if (handler == null) {
            try {
                handler = new ConfigParser(this, getClassLoader());
                handler.setNameConverter(nameConverter);
                handler.parse(in);
            } catch (ParseError e) {
                getExceptionListener().exceptionThrown(e);
            }
        }
        return handler.getObject();
    }

    /**
     * Sets the owner of this decoder to <code>owner</code>.
     *
     * @param owner The owner of this decoder.
     *
     * @see #getOwner
     */
    public void setOwner(Object owner) {
        this.owner = owner;
    }

    /**
     * Gets the owner of this decoder.
     *
     * @return The owner of this decoder.
     *
     * @see #setOwner
     */
    public Object getOwner() {
        return owner;
    }

    public NameConverter getNameConverter() {
        return nameConverter;
    }

    public void setNameConverter(NameConverter nameConverter) {
        this.nameConverter = nameConverter;
        if (handler != null) {
            handler.setNameConverter(nameConverter);
        }
    }

    public ValueResolver getValueResolver() {
        return valueResolver;
    }

    public void setValueResolver(ValueResolver valueResolver) {
        this.valueResolver = valueResolver;
    }

    public ClassResolver getClassResolver() {
        return classResolver;
    }

    public void setClassResolver(ClassResolver classResolver) {
        this.classResolver = classResolver;
    }
}
