/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this file,
 * You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Portions created by Devexperts are Copyright (C) 2007-2012 Devexperts.
 */

package com.devexperts.evolution.utils.configencoder;

import java.beans.*;
import java.io.*;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.Proxy;
import java.util.*;

/**
 * The <code>ConfigEncoder</code> class is a complementary alternative to the <code>XMLEncoder</code> and can be used to
 * generate a textual representation of a <em>JavaBean</em> in the same way that the <code>XMLEncoder</code> can be used
 * to create XML representation of <code>JavaBeans</code>.
 *
 * @author Anton V. Shakhmin
 * @see java.beans.XMLEncoder
 * @see ConfigDecoder
 */
public class ConfigEncoder extends Encoder {

    public static final String DEFAULT_ENCODING = "UTF-8";
    private static final Object[] NO_OBJECTS = new Object[]{};
    private Writer out;
    private Object owner;
    private boolean internal = false;
    private Map<Object, ValueData> valueToExpression = new IdentityHashMap<Object, ValueData>();
    private Map<Object, ArrayList<Statement>> targetToStatementList =
        new IdentityHashMap<Object, ArrayList<Statement>>();
    private NameGenerator nameGenerator = new NameGenerator();
    private NameConverter nameConverter = new NothingNameConverter();
    private AliasTable aliases = new AliasTable();
    private StringBuilder outString = new StringBuilder();


    private static class ValueData {
        public int refs = 0;
        public boolean marked = false; // Marked -> refs > 0 unless ref was a target.
        /**
         * Statement, which must be later than current exp in target statement list. Not null only for referenced values
         * and used only once, during marking of value.
         */
        public Statement beforeStatement;
        public String name = null;
        public Expression exp = null;
        public boolean processedAlready = false;
    }

    /**
     * Creates a new output stream for sending <em>JavaBeans</em> to the stream <code>out</code> using UTF-8 encoding.
     *
     * @param out The stream to which the representation of the objects will be sent.
     *
     * @see ConfigDecoder#ConfigDecoder(InputStream)
     */
    public ConfigEncoder(OutputStream out) {

        try {
            Writer w = new BufferedWriter(new OutputStreamWriter(out, DEFAULT_ENCODING));
            this.out = w;
        } catch (UnsupportedEncodingException e) {
            // UTF-8 shall be always supported
            throw new IllegalStateException("Default encoding (" + DEFAULT_ENCODING + ") is unsupported", e);
        }
    }

    public ConfigEncoder(Writer out) {
        this.out = out;
    }

    /**
     * Sets the owner of this encoder to <code>owner</code>.
     *
     * @param owner The owner of this encoder.
     *
     * @see #getOwner
     */
    public void setOwner(Object owner) {
        this.owner = owner;
        writeExpression(new Expression(this, "getOwner", NO_OBJECTS));
    }

    /**
     * Gets the owner of this encoder.
     *
     * @return The owner of this encoder.
     *
     * @see #setOwner
     */
    public Object getOwner() {
        return owner;
    }

    public NameConverter getNameConverter() {
        return nameConverter;
    }

    public void setNameConverter(NameConverter nameConverter) {
        this.nameConverter = nameConverter;
    }

    /**
     * Write an representation of the specified object to the output.
     *
     * @param o The object to be written to the stream.
     *
     * @see ConfigDecoder#readObject
     */
    @Override
    public void writeObject(Object o) {
        // The purpose of internal flag is to split the writing of the first object which we want to encode
        // and the writing of the objects which are included in first object (properties and so on)
        if (internal) {
            super.writeObject(o);
        } else {
            writeStatement(new Statement(this, "writeObject", new Object[]{o}));
        }
    }

    @Override
    public PersistenceDelegate getPersistenceDelegate(Class<?> type) {
        PersistenceDelegate delegate = super.getPersistenceDelegate(type);
        if (type != null && type.isEnum() && delegate.getClass().equals(DefaultPersistenceDelegate.class)) {
            return EnumPersistenceDelegate.getInstance();
        }
        return delegate;
    }

    private ArrayList<Statement> statementList(Object target) {
        ArrayList<Statement> list = targetToStatementList.get(target);
        if (list == null) {
            list = new ArrayList<Statement>();
            targetToStatementList.put(target, list);
        }
        return list;
    }

    /**
     * Marks objects in order to write references to another objects.
     *
     * @param o             The object to be marked.
     * @param isArgument    If the object o is argument then isArgument is true, if target then false.
     * @param fromStatement statement in which object o is used as an argument. Null if isArgument false.
     */
    private void mark(Object o, boolean isArgument, Statement fromStatement) {
        if (o == null || o == this) {
            return;
        }
        ValueData d = getValueData(o);
        Expression exp = d.exp;
        // Do not mark liternal strings. Other strings, which might,
        // for example, come from resource bundles should still be marked.
        if (o.getClass() == String.class && exp == null) {
            return;
        }

        // Bump the reference counts of all arguments
        if (isArgument) {
            d.refs++;
            /**
             * lukashevich: There was a serious bug in XMLEncoder(yes, ConfigEncoder was copied from it) with storing of
             * objects referencing each other (Bug ID: 6921644). The main problem was in incorrect order of statements -
             * when some object refers to another, which is not actually in statements list of its target (this may
             * happen when this object is a default value for some parent object) it breaks the final output to stream.
             * This issue is fixed in java7, but not in best way. They fixes the final output. Here we fix the problem
             * by correct arrangement of statements in list - those statements that is referenced must be on higher than
             * statements that refers to same value. In example:
             * Obj
             *  |
             *  |--- A1 - default value for Obj
             *  |
             *  |--- A2
             *        |
             *        |--- B1
             *              |
             *              |--- A1
             *
             *  Object A1 is default for root Obj. And B1 has the same A1 object. In old version the result
             *  statement list for Obj will be "[A2, A1]". Because A1 is marked for write only when traversing A2
             *  subtree(Obj -> A1 is default, so it won't be written in normal way, but other objects can refers to it).
             *  So, the result statement will break the output, because in this case A1 should be stored to stream
             *  before A2.
             *
             *  With this fix the result statement list for Obj will be "[A1, A2].
             */
            markReferencedValue(fromStatement.getTarget(), d);
        }
        if (d.marked) {
            return;
        }
        d.marked = true;
        Object target = exp.getTarget();
        if (!(target instanceof Class)) {
            List<Statement> statementList = statementList(target);
            // lukashevich: this is key point. If current value is referred by someone or current value has referred
            // nodes in it's subtree then this statement must be inserted before marked object. In this case it will
            // be written to file before nodes which have references to this value(or to some value in subtree).
            if (d.beforeStatement != null && !statementList.isEmpty()) {
                int index = statementList.indexOf(d.beforeStatement);
                if (index < 0) {
                    index = statementList.size();
                }
                statementList.add(index, exp);
            } else {
                statementList.add(exp);
            }
            // Pending: Why does the reference count need to
            // be incremented here?
            d.refs++;
        }
        mark(exp);
    }

    /**
     * Makes additional mark for referenced value to prevent incorrect order of statements in target list. This method
     * find actually node that is must to be additionally marked. The main idea here is to find shared root node for
     * referencedValue and the statement node. And in statements list of this shared root statement for referenced value
     * OR its subtree(if value is deep in subtree) must be inserted before the target statement(or its subtree
     * statement).
     *
     * @param statementTarget     target of statement which uses referencedValue as an argument
     * @param referencedValueData value which is referenced (used as an argument)
     */
    private void markReferencedValue(Object statementTarget, ValueData referencedValueData) {
        Object referencedValueTarget = referencedValueData.exp.getTarget();
        // if targets are equal it means that referenced value is on the same tree level and we don't
        // need to make additional marks.
        if (statementTarget == referencedValueTarget) {
            return;
        }
        // We are also not interested in Class targets - it works normally without additional marking
        if (statementTarget instanceof Class || referencedValueTarget instanceof Class) {
            return;
        }
        ValueData stmValueData = getValueData(statementTarget);
        // contains all nodes from statementTarget(including) to the root of statements tree. Main root is an Class
        // object, which is not included.
        List<Object> pathToRoot = null;
        while (stmValueData != null && stmValueData.exp != null) {
            Object nextTarget = stmValueData.exp.getTarget();
            if (nextTarget == null || nextTarget instanceof Class) {
                break;
            }
            // lazy initialization is needed, because in most cases we have case with simple and plain
            // structure (referenced node in the same level with statement target)
            if (pathToRoot == null) {
                pathToRoot = new ArrayList<Object>();
                // add initial statement target - it the first node in path to root
                pathToRoot.add(statementTarget);
            }
            pathToRoot.add(nextTarget);
            stmValueData = getValueData(nextTarget);
        }
        // if path to root is empty it means that statementTarget is the top level node or its ancestors
        // is not interested for us(Class object node) - nothing to mark.
        if (pathToRoot == null) {
            return;
        }
        while (referencedValueData != null && referencedValueData.exp != null && !referencedValueData.marked) {
            referencedValueTarget = referencedValueData.exp.getTarget();
            if (referencedValueTarget == null || referencedValueTarget instanceof Class) {
                return;
            }
            // the same reason as in the beginning
            if (referencedValueTarget == statementTarget) {
                return;
            }
            int index = identityIndexOf(pathToRoot, referencedValueTarget);
            if (index >= 0) {
                // in this place pathToRoot will always contains 2 or more elements. When shared target is found
                // we take previous element, because we need the neighbor of reverencedValue in target's statement list.
                // Referenced value (or it's parent) must be inserted before this neighbor.
                referencedValueData.beforeStatement = getValueData(pathToRoot.get(index - 1)).exp;
                return;
            }
            // switch to next node
            referencedValueData = getValueData(referencedValueTarget);
        }
    }

    /**
     * Return index of first <code>value</code> occurrence in <code>list</code> matched by identity
     * @param list - a list to be inspected
     * @param value - a value to be found
     * @return index of value in list if found; -1 otherwise
     */
    private static <E> int identityIndexOf(List<E> list, E value) {
        if (value == null) {
            return -1;
        }
        for (int i = 0; i < list.size(); i++) {
            if (list.get(i) == value) {
                return i;
            }
        }
        return -1;
    }

    /**
     * Marks every object in the argument list of a statement.
     *
     * @param stm The statement which arguments to be marked.
     */
    private void mark(Statement stm) {
        Object[] args = stm.getArguments();
        for (Object arg : args) {
            mark(arg, true, stm);
        }
        mark(stm.getTarget(), false, null);
    }

    /**
     * Records the Statement so that the Encoder will produce the actual output when the stream is flushed.
     * <p/>
     * This method should only be invoked within the context of initializing a persistence delegate.
     *
     * @param oldStm The statement that will be written to the stream.
     *
     * @see java.beans.PersistenceDelegate#initialize
     */
    public void writeStatement(Statement oldStm) {
        boolean internal = this.internal;
        this.internal = true;
        try {
            super.writeStatement(oldStm);
            /*
           Note we must do the mark first as we may
           require the results of previous values in
           this context for this statement.
           Test case is:
               os.setOwner(this);
               os.writeObject(this);
            */
            mark(oldStm);
            statementList(oldStm.getTarget()).add(oldStm);
        } catch (Exception e) {
            getExceptionListener().exceptionThrown(new Exception("ConfigEncoder: discarding statement " + oldStm, e));
        } finally {
            this.internal = internal;
        }
    }

    private Object invokeExpAndGetValue(Expression exp) {
        if (exp == null) {
            return null;
        }
        try {
            return exp.getValue();
        } catch (Exception e) {
            getExceptionListener().exceptionThrown(e);
            throw new RuntimeException("failed to evaluate: " + exp.toString(), e);
        }
    }

    /**
     * Records the Expression so that the Encoder will produce the actual output when the stream is flushed.
     * <p/>
     * This method should only be invoked within the context of initializing a persistence delegate or setting up an
     * encoder to read from a resource bundle.
     * <p/>
     * For more information about using resource bundles with the ConfigEncoder, see
     * http://java.sun.com/products/jfc/tsc/articles/persistence4/#i18n
     *
     * @param oldExp The expression that will be written to the stream.
     *
     * @see java.beans.PersistenceDelegate#initialize
     */
    public void writeExpression(Expression oldExp) {
        boolean internal = this.internal;
        this.internal = true;
        Object oldValue = invokeExpAndGetValue(oldExp);
        if (get(oldValue) == null || (oldValue instanceof String && !internal)) {
            getValueData(oldValue).exp = oldExp;
            super.writeExpression(oldExp);
        }
        this.internal = internal;
    }

    /**
     * This method calls <code>flush</code>, writes the closing postamble and then closes the output stream associated
     * with this stream.
     */
    public void close() {
        flush();
        try {
            out.close();
        } catch (IOException e) {
            getExceptionListener().exceptionThrown(e);
        }
    }

    private void registerObjectClass(Object o) {
        if (o == null) {
            return;
        }
        String className;
        if (o instanceof Class) {
            className = ((Class) o).getName();
        } else if (Proxy.isProxyClass(o.getClass())) {
            className = Proxy.getInvocationHandler(o).getClass().getName();
        } else if (o instanceof Field) {
            className = ((Field) o).getDeclaringClass().getName();
        } else {
            className = o.getClass().getName();
        }
        // REVIEW: CIT: Strange manipulations ...
        String packageName = Utils.extractPackageName(nameConverter.mapClassName(className));
        if (packageName.length() == 0) {
            return;
        }
        // REVIEW: CIT: why after mapping?
        if (packageName.charAt(0) == '[') {
            if (o instanceof Class) {
                packageName = Utils.extractPackageName(Utils.getArrayComponentClass((Class) o).getName());
            } else if (Proxy.isProxyClass(o.getClass())) {
                packageName = Utils.extractPackageName(Utils.getArrayComponentClass(Proxy.getInvocationHandler(o)
                    .getClass()).getName());
            } else {
                packageName = Utils.extractPackageName(Utils.getArrayComponentClass(o.getClass()).getName());
            }
        }
        aliases.addName(packageName);
    }

    /**
     * fills the objectNames array with aliases
     *
     * @param obj The object to be put to ArrayList
     */
    private void findAliasing(Object obj) {
        if (obj != null) {
            if ((Utils.isNanOrInfinity(obj) || !Utils.isPrimitiveWrapper(obj.getClass())) && !(obj instanceof String)) {
                if (!(obj instanceof Class) /*&& !(obj instanceof Enum)*/) {
                    ValueData d = getValueData(obj);
                    // has an object been already processed (eliminating making alias of the object which has
                    // reference)
                    if (d.processedAlready) {
                        return;
                    }
                    d.processedAlready = true;
                    registerObjectClass(obj);
                    // 'cause array has his type and the type of his elements, we process just his type
                    if (!obj.getClass().isArray()) {
                        Object[] args = d.exp.getArguments();
                        for (Object arg : args) {
                            findAliasing(arg);
                        }
                    }
                    ArrayList<Statement> statements = statementList(obj);
                    for (Statement s : statements) {
                        // if it's getter, we don't make aliases of properties
                        if (s.getClass() == Expression.class) {
                            Object value = invokeExpAndGetValue((Expression) s);
                            goThrougthGetters(value);
                        }

                        Object[] args2 = s.getArguments();
                        for (Object arg : args2) {
                            findAliasing(arg);
                        }
                    }
                } else {
                    registerObjectClass(obj);
                }
            }
        }
    }

    private void goThrougthGetters(Object obj) {
        if (obj != null) {
            if (!Utils.isPrimitiveWrapper(obj.getClass()) && !(obj instanceof String)) {
                ArrayList<Statement> statements = statementList(obj);
                for (Statement s : statements) {
                    if (s.getClass() == Expression.class) {
                        Object value = invokeExpAndGetValue((Expression) s);
                        ValueData d = getValueData(value);
                        if (d.processedAlready) {
                            return;
                        }
                        d.processedAlready = true;
                        goThrougthGetters(value);
                    }

                    Object[] args2 = s.getArguments();
                    for (Object arg : args2) {
                        findAliasing(arg);
                    }
                }
            }
        }
    }

    /**
     * This method writes out the preamble associated with the encoding if it has not been written already and then
     * writes out all of the values that been written to the stream since the last time <code>flush</code> was called.
     * After flushing, all internal references to the values that were written to this stream are cleared.
     */
    private void flush() {
        ArrayList<Statement> roots = statementList(this);
        for (Statement s : roots) {
            Object[] arguments = s.getArguments();
            if (arguments.length > 0) {
                findAliasing(arguments[0]);
            }
        }
        boolean hasAnObjectBeenAlready = false;
        for (Statement s : roots) {
            if ("writeObject".equals(s.getMethodName())) {
                if (hasAnObjectBeenAlready) {
                    write(",");
                }
                outputValue(s.getArguments()[0], this);
                hasAnObjectBeenAlready = true;
            } else {
                outputStatement(s, this);
            }
        }
        try {
            aliases.write(out);
            out.write(outString.toString());
            out.flush();
        } catch (IOException e) {
            getExceptionListener().exceptionThrown(e);
        }
        clear();
    }

    private void clear() {
        //super.clear();
        nameGenerator.clear();
        valueToExpression.clear();
        targetToStatementList.clear();
        aliases.clear();
    }

    private ValueData getValueData(Object o) {
        ValueData d = valueToExpression.get(o);
        if (d == null) {
            d = new ValueData();
            valueToExpression.put(o, d);
        }
        return d;
    }

    private void writeQuotedCharacters(String s) {
        int n = s.length();
        for (int i = 0; i < n; i++) {
            char c = s.charAt(i);
            switch (c) {
            case '\'':
                write("\\\'");
                break;
            case '"':
                write("\\\"");
                break;
            case '\r':
                write("\\r");
                break;
            case '\n':
                write("\\n");
                break;
            case '\\':
                write("\\\\");
                break;
            default:
                write(c);
            }
        }
    }

    private void write(String exp) {
        outString.append(exp);
    }

    private void write(char c) {
        outString.append(c);
    }

    //private void writeAliasing(String str) {
    //    aliasingString = aliasingString + str;
    //}

    private String useAlias(String originalClassName) {
        String mappedClassName = nameConverter.mapClassName(originalClassName);
        return aliases.getClassAlias(mappedClassName);
    }

    private String useAlias(Class clazz) {
        return useAlias(clazz.getName());
    }

    /**
     * Writes primitives to the stream.
     *
     * @param value The primitive to be written to the file.
     * @param outer The outer object
     */
    private void outputValue(Object value, Object outer) {
        if (value == null) {
            write("null");
            return;
        }

        if (value instanceof Class) {
            write("*" + useAlias(((Class) value)) + ".class");
            return;
        }

//        if (value instanceof Enum) {
//            write("&" + useAlias(value.getClass()) + "." + value);
//            return;
//        }

        ValueData d = getValueData(value);
        if (d.exp != null) {
            Object target = d.exp.getTarget();
            String methodName = d.exp.getMethodName();

            if (target == null || methodName == null) {
                throw new NullPointerException((target == null ? "target" : "methodName") + " should not be null");
            }

            if (target instanceof Field && methodName.equals("get")) {
                Field f = (Field) target;
                write("*" + useAlias(f.getDeclaringClass()) + "." + f.getName());
                return;
            }

            Class primitiveType = Utils.primitiveTypeFor(value.getClass());
            if (primitiveType != null && target == value.getClass() && methodName.equals("new")) {
                if (primitiveType == Byte.TYPE) {
                    value = value.toString() + "b";
                } else if (primitiveType == Boolean.TYPE) {
                    value = (Boolean) value ? "<T" : "<F";
                } else if (primitiveType == Character.TYPE) {
                    write("'");
                    writeQuotedCharacters(value.toString());
                    write("'");
                    return;
                } else if (primitiveType == Short.TYPE) {
                    value = value.toString() + "s";
                } else if (primitiveType == Integer.TYPE) {
                    value = value.toString();
                } else if (primitiveType == Long.TYPE) {
                    value = value.toString() + "l";
                } else if (primitiveType == Float.TYPE) {
                    if (((Float) value).isNaN()) {
                        write("*" + useAlias(value.getClass()) + "." + "NaN");
                        return;
                    }
                    if ((Float) value == Float.POSITIVE_INFINITY) {
                        write("*" + useAlias(value.getClass()) + "." + "POSITIVE_INFINITY");
                        return;
                    }
                    if ((Float) value == Float.NEGATIVE_INFINITY) {
                        write("*" + useAlias(value.getClass()) + "." + "NEGATIVE_INFINITY");
                        return;
                    }
                    value = value.toString() + "f";
                } else if (primitiveType == Double.TYPE) {
                    if (((Double) value).isNaN()) {
                        write("*" + useAlias(value.getClass()) + "." + "NaN");
                        return;
                    }
                    if ((Double) value == Double.POSITIVE_INFINITY) {
                        write("*" + useAlias(value.getClass()) + "." + "POSITIVE_INFINITY");
                        return;
                    }
                    if ((Double) value == Double.NEGATIVE_INFINITY) {
                        write("*" + useAlias(value.getClass()) + "." + "NEGATIVE_INFINITY");
                        return;
                    }
                    value = value.toString() + "d";
                }
                write(value.toString());
                return;
            }
        } else if (value instanceof String) {
            write("\"");
            writeQuotedCharacters((String) value);
            write("\"");
            return;
        }

        if (d.name != null) {
            write("@" + d.name);
            return;
        }

        outputStatement(d.exp, outer);
    }

    /**
     * Writes expressions and statements to the stream.
     *
     * @param exp   The expression to be written to the file.
     * @param outer The outer object
     */
    private void outputStatement(Statement exp, Object outer) {
        Object target = exp.getTarget();
        String methodName = exp.getMethodName();
        if (target instanceof Class) {
            methodName = nameConverter.mapMethodName(((Class) target).getName(), methodName);
        } else {
            methodName = nameConverter.mapMethodName(target.getClass().getName(), methodName);
        }
        if (target == null || methodName == null) {
            throw new NullPointerException((target == null ? "target" : "methodName") + " should not be null");
        }

        Object[] args = exp.getArguments();
        boolean expression = exp.getClass() == Expression.class;
        Object value = (expression) ? invokeExpAndGetValue((Expression) exp) : null;

        String attributes = "";
        String refId = "";

        ValueData d = getValueData(value);
        if (expression) {
            if (d.refs > 1) {
                String instanceName = nameGenerator.instanceName(value);
                d.name = instanceName;
                refId = ":" + instanceName;
            }
        }
        String signNew;
        // Special cases for targets.
        if (target == outer) {
            // Do nothing
        } else if (target == Array.class && methodName.equals("newInstance")) {
            signNew = "#";
            Class componentClass = (Class) args[0];
            String squareBrackets = "";
            if (componentClass.isArray()) {
                squareBrackets = calculateSquareBrackets(componentClass);
                componentClass = Utils.getArrayComponentClass(componentClass);
            }
            attributes += signNew + useAlias(componentClass);
            attributes += "[" + args[1].toString() + "]" + squareBrackets;
            args = NO_OBJECTS;
        } else if (target.getClass() == Class.class) {
            if (methodName.equals("new") || methodName.equals("newInstance")) {
                signNew = "#";
            } else {
                signNew = "!";
            }
            attributes = attributes + signNew + useAlias(((Class) target));
        } else {
            d.refs = 2;
            outputValue(target, outer);
            outputValue(value, outer);
            return;
        }
        // Special cases for methods.
        // IndexedSetter
        if ((!expression && methodName.equals("set") && args.length == 2 && args[0] instanceof Integer)) {
            attributes = attributes + args[0].toString() + "=";
            args = (args.length == 1) ? NO_OBJECTS : new Object[]{args[1]};
            // IndexedGetter
        } else if ((expression && methodName.equals("get") && args.length == 1 && args[0] instanceof Integer)) {
            attributes = attributes + args[0].toString() + ":";
            args = (args.length == 1) ? NO_OBJECTS : new Object[]{args[1]};
            // property's getter and setter
        } else if ((!expression && methodName.startsWith("set") && args.length == 1)
            || (expression && methodName.startsWith("get") && args.length == 0))
        {
            ArrayList<Statement> statements = statementList(value);
            attributes = attributes + Introspector.decapitalize(methodName.substring(3));
            // if there is just one primitive or object in the context of the current property
            if (args.length == 1 && statements.size() == 0) {
                attributes = attributes + "=";
                write(attributes);
                outputValue(args[0], null);
                return;
                // if there is just one property or method in the context of the current property
            } else if (args.length == 0 && statements.size() == 1 && d.refs < 2) {
                // if the property in the context of the current property is not indexed
                Statement statement = statements.get(0);
                if (!((statement.getMethodName().equals("set") && statement.getArguments().length == 2
                        && statement.getArguments()[0] instanceof Integer)
                        || (statement.getMethodName().equals("get")
                            && statement.getArguments().length == 1
                            && statement.getArguments()[0] instanceof Integer)))
                {
                    attributes = attributes + ".";
                    write(attributes);
                    outputStatement(statements.iterator().next(), value);
                    return;
                }
                // if there are many properties or methods or primitives or objects in the context of the current
                // property
            }
            // methods
        } else if (!methodName.equals("new") && !methodName.equals("newInstance")) {
            if (methodName.equals("add")) {
                attributes = attributes + "+";
            } else if (methodName.equals("put")) {
                attributes = attributes + "^";
            } else if (target.getClass() == Class.class) {
                attributes = attributes + "." + methodName;
            } else {
                attributes = attributes + "!" + methodName;
            }
        }
        //reference to an object is added despite the object is an array, property or returned by a method call
        attributes = attributes + refId;

        ArrayList<Statement> statements = statementList(value);
        if (args.length == 0 && statements.size() == 0) {
            write(attributes);
            return;
        }

        write(attributes);

        if (args.length != 0) {
            // we don't write '(' and ')' for array elements
            if (expression || !methodName.equals("set") || args.length != 1) {
                write("(");
            }
            if (args.length == 1) {
                outputValue(args[0], null);
            } else {
                outputValue(args[0], null);
                for (int i = 1; i < args.length; i++) {
                    Object arg = args[i];
                    write(",");
                    outputValue(arg, null);
                }
            }
            if (expression || !methodName.equals("set") || args.length != 1) {
                write(")");
            }
        }

        if (statements.size() != 0) {
            write("{");
            if (statements.size() == 1) {
                outputStatement(statements.iterator().next(), value);
            } else {
                outputStatement(statements.iterator().next(), value);
                Iterator<Statement> it = statements.iterator();
                it.next();
                while (it.hasNext()) {
                    Statement s = it.next();
                    write(",");
                    outputStatement(s, value);
                }
            }
            write("}");
        }
    }

    private String calculateSquareBrackets(Class aType) {
        String res = "";
        while (aType.isArray()) {
            res = res + "[]";
            aType = aType.getComponentType();
        }
        return res;
    }

    private static class NameGenerator {
        private Map<Object, String> valueToName;
        private Map<String, Integer> nameToCount;

        public NameGenerator() {
            valueToName = new IdentityHashMap<Object, String>();
            nameToCount = new HashMap<String, Integer>();
        }

        /**
         * Clears the name cache. Should be called to near the end of the encoding cycle.
         */
        public void clear() {
            valueToName.clear();
            nameToCount.clear();
        }

        // Returns the root name of the class.
        public static String unqualifiedClassName(Class<? extends Object> type) {
            if (type.isArray()) {
                return unqualifiedClassName(type.getComponentType()) + "Array";
            }
            String name = type.getName();
            return name.substring(name.lastIndexOf('.') + 1);
        }

        // Returns a String which capitalizes the first letter of the string.
        public static String capitalize(String name) {
            if (name == null || name.length() == 0) {
                return name;
            }
            return name.substring(0, 1).toUpperCase() + name.substring(1);
        }

        /**
         * Returns a unique string which identifies the object instance. Invocations are cached so that if an object has
         * been previously passed into this method then the same identifier is returned.
         *
         * @param instance object used to generate string
         *
         * @return a unique string representing the object
         */
        public String instanceName(Object instance) {
            if (instance == null) {
                return "null";
            }
            if (instance instanceof Class) {
                return unqualifiedClassName((Class<? extends Object>) instance);
            } else {
                String result = valueToName.get(instance);
                if (result != null) {
                    return result;
                }
                Class<? extends Object> type = instance.getClass();
                String className = unqualifiedClassName(type);

                Object size = nameToCount.get(className);
                int instanceNumber = (size == null) ? 0 : (Integer) size + 1;
                nameToCount.put(className, instanceNumber);

                result = className + instanceNumber;
                valueToName.put(instance, result);
                return result;
            }
        }
    }
}
