/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this file,
 * You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Portions created by Devexperts are Copyright (C) 2007-2012 Devexperts.
 */

package com.devexperts.evolution.utils.configencoder;

import java.util.Hashtable;
import java.util.ArrayList;
import java.beans.Expression;
import java.io.*;
import java.lang.reflect.Array;
import java.lang.reflect.Field;

/**
 * ConfigParser class is used to parse textual documents created by ConfigEncoder
 *
 * @author Anton V. Shakhmin
 * @see ConfigEncoder
 * @see ConfigDecoder
 */
public class ConfigParser {
    // todo: any reason why it is synchronized Hashtable and not HashMap here?
    private final Hashtable<String, MutableExpression> environment = new Hashtable<String, MutableExpression>();
    private final ArrayList<MutableExpression> expStack = new ArrayList<MutableExpression>();
    private final ConfigDecoder is;
    private final ArrayList<String> aliases = new ArrayList<String>();

    private ClassLoader ldr;
    private int itemsRead = 0;
    private String strIn;
    private int marker = 0;
    private ArrayList<Object> objects;
    private Lexeme curLexeme = null;
    private NameConverter nameConverter = new NothingNameConverter();

    public ConfigParser(ConfigDecoder is) {
        if (is == null) {
            throw new NullPointerException("ConfigDecoder");
        }
        this.is = is;
    }

    /* loader can be null */
    public ConfigParser(ConfigDecoder is, ClassLoader loader) {
        this(is);
        this.ldr = loader;
    }

    // returns true if ch equals to current symbol
    private boolean isCurSym(char ch) {
        return strIn.charAt(marker) == ch;
    }

    // returns n symbols following marker
    private String curSym(int number) {
        return strIn.substring(marker, marker + number);
    }

    // returns n symbols following marker + offset
    private String nextSymOffset(int offset, int number) {
        return strIn.substring(marker + offset, marker + offset + number);
    }

    // returns true if the symbol following marker + offset equals ch
    private boolean isNextSymOffset(int offset, char ch) {
        return strIn.charAt(marker + offset) == ch;
    }

    // returns the string beginning from marker and to the end of the InputString
    private String restOfInStr() {
        return strIn.substring(marker);
    }

    // returns the length of the string beginning from marker
    private int lengthOfRest() {
        return restOfInStr().length();
    }

    public void parse(InputStream in) throws ParseError {
        try {
            BufferedReader bufIn = new BufferedReader(new InputStreamReader(in, ConfigEncoder.DEFAULT_ENCODING));
            strIn = bufIn.readLine();
            if (strIn == null) {
                throw new ParseError("Empty file or stream");
            }
            curLexeme = new Lexeme();
            readCurLexeme();
            objects = readExpressions();
        } catch (IOException e) {
            throw new ParseError("File or stream reading error");
            //e.printStackTrace();
        }
    }

    private void readCurLexeme() throws ParseError {
        if (lengthOfRest() > 0) {
            if (isCurSym('#')) {
                marker++;
                curLexeme.setType(Lexeme.NEW_INSTANCE_SIGN);
                curLexeme.setValue("#");
                return;
            }
            if (isCurSym('(')) {
                marker++;
                curLexeme.setType(Lexeme.LEFT_PARENTHESIS);
                curLexeme.setValue("(");
                return;
            }
            if (isCurSym(')')) {
                marker++;
                curLexeme.setType(Lexeme.RIGHT_PARENTHESIS);
                curLexeme.setValue(")");
                return;
            }
            if (isCurSym('{')) {
                marker++;
                curLexeme.setType(Lexeme.LEFT_BRACE);
                curLexeme.setValue("{");
                return;
            }
            if (isCurSym('}')) {
                marker++;
                curLexeme.setType(Lexeme.RIGHT_BRACE);
                curLexeme.setValue("}");
                return;
            }
            if (isCurSym(',')) {
                marker++;
                curLexeme.setType(Lexeme.COMMA);
                curLexeme.setValue(",");
                return;
            }
            if (isCurSym('=')) {
                marker++;
                curLexeme.setType(Lexeme.EQUALS_SIGN);
                curLexeme.setValue("=");
                return;
            }
            if (isCurSym('@')) {
                marker++;
                curLexeme.setType(Lexeme.OBJECT_REF);
                curLexeme.setValue("@");
                return;
            }
            if (isCurSym('%')) {
                marker++;
                curLexeme.setType(Lexeme.ALIAS_SIGN);
                curLexeme.setValue("%");
                return;
            }
            //boolean
            if (isCurSym('<')) {
                curLexeme.setType(Lexeme.BOOLEAN);
                String bool = "" + curSym(2);
                if (!(isNextSymOffset(1, 'F') || isNextSymOffset(1, 'T'))) {
                    throw new ParseError("Incorrect of boolean at position " + marker
                        + ". Expected '<F' or '<T', but was '" + bool + "'  t \'");
                }
                curLexeme.setValue(bool);
                marker += 2;
                return;
            }
            //numbers
            if (isCurSym('-') || Character.isDigit(curSym(1).charAt(0))) {
                readNumberLexeme();
                return;
            }
            //char
            if (isCurSym('\'')) {
                int i = 1;
                if (i >= lengthOfRest()) {
                    throw new ParseError("Incorrect char: the end of the file is reached, but there is no sign \'");
                }
                StringBuilder num = new StringBuilder("");
                String cur = nextSymOffset(i, 1);
                char c = nextSymOffset(i, 1).charAt(0);
                if (c == '\\') {
                    if (i + 1 < lengthOfRest()) {
                        i++;
                        char c2 = nextSymOffset(i, 1).charAt(0);
                        if (c2 == '\"') {
                            cur = "\"";
                        } else if (c2 == '\'') {
                            cur = "\'";
                        } else if (c2 == 'r') {
                            cur = "\r";
                        } else if (c2 == '\\') {
                            cur = "\\";
                        } else if (c2 == 'n') {
                            cur = "\n";
                        } else {
                            throw new ParseError("Incorrect char: the sign " + nextSymOffset(i, 1)
                                + " at position " + (marker + i) + " should not follow sign \\");
                        }
                    }
                }
                num.append(cur);
                i++;
                if (i >= lengthOfRest()) {
                    throw new ParseError("Incorrect char: the end of the file is reached, but there is no sign \'");
                }
                if (!isNextSymOffset(i, '\'')) {
                    throw new ParseError("Incorrect char at position " + (marker + i));
                }
                marker = marker + i + 1;
                curLexeme.setType(Lexeme.CHAR);
                curLexeme.setValue(num.toString());
                return;
            }
            //String
            if (isCurSym('"')) {
                int i = 1;
                if (i >= lengthOfRest()) {
                    throw new ParseError("Incorrect String: the end of the file is reached, but there is no sign \"");
                }
                StringBuilder num = new StringBuilder("");
                if (!isNextSymOffset(i, '"')) {
                    boolean isEnd = false;
                    while (i < lengthOfRest() && !isEnd) {
                        String cur = nextSymOffset(i, 1);
                        char c = nextSymOffset(i, 1).charAt(0);
                        if (c == '\\') {
                            if (i + 1 < lengthOfRest()) {
                                i++;
                                char c2 = nextSymOffset(i, 1).charAt(0);
                                if (c2 == '\"') {
                                    cur = "\"";
                                } else if (c2 == '\'') {
                                    cur = "\'";
                                } else if (c2 == 'r') {
                                    cur = "\r";
                                } else if (c2 == '\\') {
                                    cur = "\\";
                                } else if (c2 == 'n') {
                                    cur = "\n";
                                } else {
                                    throw new ParseError("Incorrect String: the sign " + nextSymOffset(i, 1)
                                        + " at position " + (marker + i) + " should not follow sign \\");
                                }
                            }
                        }
                        num.append(cur);
                        i++;
                        if (i >= lengthOfRest()) {
                            throw new ParseError(
                                "Incorrect String: the end of the file is reached, but there is no sign \"");
                        }
                        if (isNextSymOffset(i, '"')) {
                            isEnd = true;
                        }
                    }
                }
                marker = marker + i + 1;
                curLexeme.setType(Lexeme.STRING);
                curLexeme.setValue(num.toString());
                return;
            }
            //null
            if (lengthOfRest() > 3 && curSym(4).equals("null")) {
                marker = marker + 4;
                curLexeme.setType(Lexeme.NULL);
                curLexeme.setValue("null");
                return;
            }
            //Name
            if (Character.isJavaIdentifierStart(curSym(1).charAt(0))) {
                readName();
                return;
            }
            //':'
            if (isCurSym(':')) {
                marker++;
                curLexeme.setType(Lexeme.COLON);
                curLexeme.setValue(":");
                return;
            }
            // dimension in array type
            if (lengthOfRest() > 1 && curSym(2).equals("[]")) {
                int i = 2;
                int count = 1;
                while (i + 1 < lengthOfRest() && nextSymOffset(i, 2).equals("[]")) {
                    i = i + 2;
                    count++;
                }
                marker = marker + i;
                curLexeme.setType(Lexeme.DIMENSION_IN_ARRAY_TYPE);
                curLexeme.setValue(String.valueOf(count));
                return;
            }
            //array dimension
            if (lengthOfRest() > 2 && isCurSym('[')) {
                marker++;
                int i = 0;
                String num = "";
                while (i < lengthOfRest() && !isNextSymOffset(i, ']')) {
                    num = num + nextSymOffset(i, 1);
                    i++;
                }
                marker = marker + i + 1;
                curLexeme.setType(Lexeme.ARRAY_DIMENSION);
                curLexeme.setValue(num);
                return;
            }
            //instance of class or field
            if (isCurSym('*')) {
                marker++;
                curLexeme.setType(Lexeme.CLASS_OR_FIELD);
                curLexeme.setValue("*");
                return;
            }
            //'+'
            if (isCurSym('+')) {
                marker++;
                curLexeme.setType(Lexeme.PLUS_SIGN);
                curLexeme.setValue("+");
                return;
            }
            //'^'
            if (isCurSym('^')) {
                marker++;
                curLexeme.setType(Lexeme.DEGREE_SIGN);
                curLexeme.setValue("^");
                return;
            }
            //'!'
            if (isCurSym('!')) {
                marker++;
                curLexeme.setType(Lexeme.EXCLAMATION_MARK);
                curLexeme.setValue("!");
                return;
            }
            //enum
            if (isCurSym('&')) {
                marker++;
                curLexeme.setType(Lexeme.ENUM);
                curLexeme.setValue("&");
                return;
            }
            throw new ParseError(
                "The lexeme beginning with \"" + curSym(1) + "\" at position " + marker + " doesn't exist");
        } else {
            curLexeme = null;
        }
    }

    private void readName() {
        int i = 0;
        String name = "";
        while (i < lengthOfRest() && !isNextSymOffset(i, ':') && !isNextSymOffset(i, '(')
            && !isNextSymOffset(i, '{') && !isNextSymOffset(i, ',') && !isNextSymOffset(i, ')')
            && !isNextSymOffset(i, '}') && !isNextSymOffset(i, '=') && !isNextSymOffset(i, '[')
            && !isNextSymOffset(i, '%'))
        {
            name = name + nextSymOffset(i, 1);
            i++;
        }
        marker = marker + i;
        curLexeme.setType(Lexeme.NAME);
        curLexeme.setValue(name);
    }

    private void readNumberLexeme() throws ParseError {
        int i;
        String num = "";
        if (isCurSym('-')) {
            i = 1;
            num = "-";
        } else {
            i = 0;
        }
        while (i < lengthOfRest() && (Character.isDigit(nextSymOffset(i, 1).charAt(0)))) {
            num = num + nextSymOffset(i, 1);
            i++;
        }
        marker = marker + i;
        i = 0;
        if (marker < strIn.length()) {
            if (isCurSym('b')) {
                marker++;
                curLexeme.setType(Lexeme.BYTE);
                curLexeme.setValue(num);
                return;
            }
            if (isCurSym('s')) {
                marker++;
                curLexeme.setType(Lexeme.SHORT);
                curLexeme.setValue(num);
                return;
            }
            if (isCurSym('l')) {
                marker++;
                curLexeme.setType(Lexeme.LONG);
                curLexeme.setValue(num);
                return;
            }
            if (isCurSym('f')) {
                marker++;
                curLexeme.setType(Lexeme.FLOAT);
                curLexeme.setValue(num);
                return;
            }
            if (isCurSym('d')) {
                marker++;
                curLexeme.setType(Lexeme.DOUBLE);
                curLexeme.setValue(num);
                return;
            }
            if (isCurSym('.')) {
                num = num + curSym(1);
                marker++;
                while (i < lengthOfRest() && (Character.isDigit(nextSymOffset(i, 1).charAt(0))
                    || nextSymOffset(i, 1).charAt(0) == 'E' || nextSymOffset(i, 1).charAt(0) == '-'))
                {
                    num = num + nextSymOffset(i, 1);
                    i++;
                }
                if (i >= lengthOfRest()) {
                    throw new ParseError(
                        "Incorrect float or double: the end of the file is reached, but there is no sign f or d");
                }
                marker = marker + i;
                if (isCurSym('f')) {
                    marker++;
                    curLexeme.setType(Lexeme.FLOAT);
                    curLexeme.setValue(num);
                    return;
                } else if (isCurSym('d')) {
                    marker++;
                    curLexeme.setType(Lexeme.DOUBLE);
                    curLexeme.setValue(num);
                    return;
                } else {
                    throw new ParseError("Incorrect format of Float or Double at position " + marker);
                }
            }
        }
        curLexeme.setType(Lexeme.INT);
        curLexeme.setValue(num);
    }

    private ArrayList<Object> readExpressions() throws ParseError {
        ArrayList<Object> result = new ArrayList<Object>();
        if (curLexeme.getType() == Lexeme.ALIAS_SIGN) {
            readAliasDefinitions();
        }
        if (curLexeme != null && curLexeme.getType() == Lexeme.NAME && curLexeme.getValue().startsWith("owner")) {
            readProperty();
        }
        result.add(readExpression());
        while (curLexeme != null && curLexeme.getType() != Lexeme.RIGHT_BRACE
            && curLexeme.getType() != Lexeme.RIGHT_PARENTHESIS)
        {
            if (curLexeme.getType() == Lexeme.COMMA) {
                readCurLexeme();
            } else {
                throw new ParseError("The lexeme at position " + (marker - curLexeme.getValue().length())
                    + " must be comma rather than " + curLexeme.getValue());
            }
            result.add(readExpression());
        }
        return result;
    }

    private void readAliasDefinitions() throws ParseError {
        readCurLexeme();
        readAliasDefinition();
        while (curLexeme != null && curLexeme.getType() != Lexeme.ALIAS_SIGN) {
            if (curLexeme.getType() == Lexeme.COMMA) {
                readCurLexeme();
            } else {
                throw new ParseError("The lexeme at position " + (marker - curLexeme.getValue().length())
                    + " must be comma rather than " + curLexeme.getValue());
            }
            readAliasDefinition();
        }
        readCurLexeme();
    }

    private void readAliasDefinition() throws ParseError {
        if (curLexeme == null) {
            throw new ParseError("There is no number following sign %");
        }
        if (curLexeme.getType() != Lexeme.INT) {
            throw new ParseError("The lexeme " + curLexeme.getValue() + " at position "
                + (marker - curLexeme.getValue().length()) + " must be a number");
        }
        readCurLexeme();
        if (curLexeme == null) {
            throw new ParseError("There is no colon following number");
        }
        if (curLexeme.getType() != Lexeme.COLON) {
            throw new ParseError("The lexeme " + curLexeme.getValue() + " at position "
                + (marker - curLexeme.getValue().length()) + " must be a colon");
        }
        readCurLexeme();
        if (curLexeme == null) {
            throw new ParseError("There is no package name following colon");
        }
        if (curLexeme.getType() != Lexeme.NAME) {
            throw new ParseError("The lexeme " + curLexeme.getValue() + " at position "
                + (marker - curLexeme.getValue().length()) + " must be a package name");
        }
        aliases.add(curLexeme.getValue());
        readCurLexeme();
    }

    private Object readExpression() throws ParseError {
        if (curLexeme.getType() == Lexeme.NEW_INSTANCE_SIGN) {
            return readNewInstance();
        }
        if (curLexeme.getType() == Lexeme.OBJECT_REF) {
            return readObjectRef();
        }
        if (curLexeme.getType() == Lexeme.BOOLEAN) {
            boolean b = curLexeme.getValue().equals("<T");
            readCurLexeme();
            return b;
        }
        if (curLexeme.getType() == Lexeme.BYTE) {
            byte by;
            try {
                by = Byte.parseByte(curLexeme.getValue());
            } catch (NumberFormatException e) {
                throw new ParseError("Incorrect format of byte " + curLexeme.getValue());
            }
            readCurLexeme();
            return by;
        }
        if (curLexeme.getType() == Lexeme.SHORT) {
            short sh;
            try {
                sh = Short.parseShort(curLexeme.getValue());
            } catch (NumberFormatException e) {
                throw new ParseError("Incorrect format of short " + curLexeme.getValue());
            }
            readCurLexeme();
            return sh;
        }
        if (curLexeme.getType() == Lexeme.INT) {
            int i;
            try {
                i = Integer.parseInt(curLexeme.getValue());
            } catch (NumberFormatException e) {
                throw new ParseError("Incorrect format of integer " + curLexeme.getValue());
            }
            readCurLexeme();
            return i;
        }
        if (curLexeme.getType() == Lexeme.LONG) {
            long l;
            try {
                l = Long.parseLong(curLexeme.getValue());
            } catch (NumberFormatException e) {
                throw new ParseError("Incorrect format of long " + curLexeme.getValue());
            }
            readCurLexeme();
            return l;
        }
        if (curLexeme.getType() == Lexeme.FLOAT) {
            float f;
            try {
                f = Float.parseFloat(curLexeme.getValue());
            } catch (NumberFormatException e) {
                throw new ParseError("Incorrect format of float " + curLexeme.getValue());
            }
            readCurLexeme();
            return f;
        }
        if (curLexeme.getType() == Lexeme.DOUBLE) {
            double d;
            try {
                d = Double.parseDouble(curLexeme.getValue());
            } catch (NumberFormatException e) {
                throw new ParseError("Incorrect format of double " + curLexeme.getValue());
            }
            readCurLexeme();
            return d;
        }
        if (curLexeme.getType() == Lexeme.CHAR) {
            char c = curLexeme.getValue().charAt(0);
            readCurLexeme();
            return c;
        }
        if (curLexeme.getType() == Lexeme.STRING) {
            String s = curLexeme.getValue();
            readCurLexeme();
            return s;
        }
        if (curLexeme.getType() == Lexeme.NULL) {
            readCurLexeme();
            return null;
        }
        if (curLexeme.getType() == Lexeme.CLASS_OR_FIELD) {
            return readClassOrField();
        }
        if (curLexeme.getType() == Lexeme.ENUM) {
            return readEnum();
        }
        if (curLexeme.getType() == Lexeme.EXCLAMATION_MARK) {
            return readObjectReturnedByStaticMethod();
        }

        throw new ParseError("The expression beginning with lexeme " + curLexeme.getValue() + " at position "
            + (marker - curLexeme.getValue().length()) + " does not exist");
    }

    private Object readObjectReturnedByStaticMethod() throws ParseError {
        readCurLexeme();
        if (curLexeme == null) {
            throw new ParseError("There is no static method name following sign !");
        }
        if (curLexeme.getType() != Lexeme.NAME && curLexeme.getType() != Lexeme.ALIAS_SIGN) {
            throw new ParseError("The lexeme " + curLexeme.getValue() + " at position "
                + (marker - curLexeme.getValue().length()) + " must be the name of the static method");
        }
        String methodName;
        String className;
        if (curLexeme.getType() == Lexeme.ALIAS_SIGN) {
            String fullName = readAlias();
            className = fullName.substring(0, fullName.lastIndexOf('.'));
            methodName = fullName.substring(fullName.lastIndexOf('.') + 1);
        } else {
            if (curLexeme.getValue().lastIndexOf('.') == -1) {
                throw new ParseError("Incorrect name of the static method: " + curLexeme.getValue() + " at position "
                    + (marker - curLexeme.getValue().length()));
            } else {
                className = curLexeme.getValue().substring(0, curLexeme.getValue().lastIndexOf('.'));
                methodName = curLexeme.getValue().substring(curLexeme.getValue().lastIndexOf('.') + 1);
            }
        }
        if (methodName == null || className == null || methodName.equals("") || className.equals("")) {
            throw new ParseError("Incorrect name of the static method: " + curLexeme.getValue() + " at position "
                + (marker - curLexeme.getValue().length()));
        }
        readCurLexeme();
        String defRef = readDefRef();
        MutableExpression e = new MutableExpression(nameConverter);
        //reading of the argument list of object's constractor
        ArrayList<Object> args = argList();
        e.setTarget(classForName2(nameConverter.demapClassName(className)));
        e.setMethodName(methodName);
        if (args != null) {
            for (Object arg : args) {
                e.addArg(arg);
            }
        }
        //reading of the statement list of object or the element list of array
        e = getExpressionWithStatements(e);
        if (defRef != null) {
            environment.put(defRef, e);
        }
        return getValue(e);
    }

    private Object readObjectRef() throws ParseError {
        readCurLexeme();
        if (curLexeme == null) {
            throw new ParseError("There is no object reference name following sign @");
        }
        if (curLexeme.getType() != Lexeme.NAME) {
            throw new ParseError("The lexeme " + curLexeme.getValue() + " at position "
                + (marker - curLexeme.getValue().length()) + " must be the name of the object reference");
        }
        Expression e = environment.get(curLexeme.getValue());
        readCurLexeme();
        return getValue(e);
    }

    private Object readEnum() throws ParseError {
        readCurLexeme();
        if (curLexeme == null) {
            throw new ParseError("There is no enum name following sign &");
        }
        if (curLexeme.getType() != Lexeme.NAME && curLexeme.getType() != Lexeme.ALIAS_SIGN) {
            throw new ParseError("The lexeme " + curLexeme.getValue() + " at position "
                + (marker - curLexeme.getValue().length()) + " must be the name of enum");
        }
        String enumName;
        String className;
        if (curLexeme.getType() == Lexeme.ALIAS_SIGN) {
            String fullName = readAlias();
            className = fullName.substring(0, fullName.lastIndexOf('.'));
            enumName = fullName.substring(fullName.lastIndexOf('.') + 1);
        } else {
            if (curLexeme.getValue().lastIndexOf('.') == -1) {
                throw new ParseError("Incorrect name of enum: " + curLexeme.getValue() + " at position "
                    + (marker - curLexeme.getValue().length()));
            } else {
                className = curLexeme.getValue().substring(0, curLexeme.getValue().lastIndexOf('.'));
                enumName = curLexeme.getValue().substring(curLexeme.getValue().lastIndexOf('.') + 1);
            }
        }
        MutableExpression e = new MutableExpression(nameConverter);
        e.setTarget(classForName2(nameConverter.demapClassName(className)));
        e.setMethodName("valueOf");
        e.addArg(classForName2(nameConverter.demapClassName(className)));
        e.addArg(enumName);
        readCurLexeme();
        return getValue(e);
    }

    private Object readClassOrField() throws ParseError {
        readCurLexeme();
        if (curLexeme == null) {
            throw new ParseError("There is no class or field name following sign *");
        }
        if (curLexeme.getType() != Lexeme.NAME && curLexeme.getType() != Lexeme.ALIAS_SIGN) {
            throw new ParseError("The lexeme " + curLexeme.getValue() + " at position "
                + (marker - curLexeme.getValue().length()) + " must be the name of class or field");
        }
        String fullName = curLexeme.getValue();
        if (curLexeme.getType() == Lexeme.ALIAS_SIGN) {
            fullName = readAlias();
        } else {
            if (curLexeme.getValue().lastIndexOf('.') == -1) {
                throw new ParseError("Incorrect name of class or field "
                    + curLexeme.getValue() + " at position " + (marker - curLexeme.getValue().length()));
            }
        }
        MutableExpression e = new MutableExpression(nameConverter);
        if (fullName.substring(fullName.lastIndexOf('.') + 1).equals("class")) {
            e.setTarget(Class.class);
            e.setMethodName("forName");
            e.addArg(nameConverter.demapClassName(fullName.substring(0, fullName.lastIndexOf('.'))));
        } else {
            String className = fullName.substring(0, fullName.lastIndexOf('.'));
            String fieldName = fullName.substring(fullName.lastIndexOf('.') + 1);
            e.setTarget(classForName2(nameConverter.demapClassName(className)));
            e.setValue(getFieldValue(e.getTarget(), fieldName));
        }
        readCurLexeme();
        return getValue(e);
    }

    private void readStatements() throws ParseError {
        readStatement();
        while (curLexeme != null && curLexeme.getType() != Lexeme.RIGHT_BRACE
            && curLexeme.getType() != Lexeme.RIGHT_PARENTHESIS)
        {
            if (curLexeme.getType() == Lexeme.COMMA) {
                readCurLexeme();
            } else {
                throw new ParseError("The lexeme at position " + (marker - curLexeme.getValue().length())
                    + " must be comma rather than " + curLexeme.getValue());
            }
            readStatement();
        }
    }

    private void readStatement() throws ParseError {
        //properties
        if (curLexeme.getType() == Lexeme.NAME) {
            readProperty();
            return;
        }
        //indexed setter and getter
        if (curLexeme.getType() == Lexeme.INT) {
            readIndexedGetterOrSetter();
            return;
        }
        //methods
        if (curLexeme.getType() == Lexeme.PLUS_SIGN || curLexeme.getType() == Lexeme.DEGREE_SIGN
            || curLexeme.getType() == Lexeme.EXCLAMATION_MARK)
        {
            readMethod();
            return;
        }
        throw new ParseError("The statement beginning with lexeme " + curLexeme.getValue() + " at position "
            + (marker - curLexeme.getValue().length()) + " does not exist");
    }

    private void readMethod() throws ParseError {
        String methodName;
        if (curLexeme.getType() == Lexeme.PLUS_SIGN) {
            methodName = "add";
        } else if (curLexeme.getType() == Lexeme.DEGREE_SIGN) {
            methodName = "put";
        } else if (curLexeme.getType() == Lexeme.EXCLAMATION_MARK && curLexeme.getValue().length() > 1) {
            String lexemeValue = curLexeme.getValue();
            methodName = lexemeValue.substring(1, lexemeValue.length());
        } else {
            readCurLexeme();
            if (curLexeme == null) {
                throw new ParseError("There is no method name at the end of the file");
            }
            if (curLexeme.getType() != Lexeme.NAME) {
                throw new ParseError("The lexeme " + curLexeme.getValue() + " at position "
                    + (marker - curLexeme.getValue().length()) + " must be the name of the method");
            } else {
                methodName = curLexeme.getValue();
            }
        }
        readCurLexeme();
        String defRef = readDefRef();
        MutableExpression e = new MutableExpression(nameConverter);
        //reading of the argument list of method
        ArrayList<Object> args = argList();
        e.setTarget(eval());
        e.setMethodName(methodName);
        if (args != null) {
            for (Object arg : args) {
                e.addArg(arg);
            }
        }
        //reading of the statement list of method
        e = getExpressionWithStatements(e);
        if (defRef != null) {
            environment.put(defRef, e);
        }
        getValue(e);
    }

    private void readProperty() throws ParseError {
        String propName = curLexeme.getValue();
        String[] properties = null;
        //if there are several properties via dot (getters and setter)
        if (curLexeme.getValue().indexOf('.') != -1) {
            properties = curLexeme.getValue().split("\\.");
            if (properties.length == 1) {
                throw new ParseError("Incorrect lexeme " + curLexeme.getValue() + " at position "
                    + (marker - curLexeme.getValue().length()));
            }
            for (int i = 0; i < properties.length - 1; i++) {
                MutableExpression e = new MutableExpression(nameConverter);
                e.setProperty(properties[i]);
                e.setTarget(eval());
                expStack.add(e);
            }
            //if setter is a method
            char c = properties[properties.length - 1].charAt(0);
            if (c == '+') {
                curLexeme.setType(Lexeme.PLUS_SIGN);
            } else if (c == '^') {
                curLexeme.setType(Lexeme.DEGREE_SIGN);
            } else if (c == '!') {
                curLexeme.setType(Lexeme.EXCLAMATION_MARK);
                curLexeme.setValue(properties[properties.length - 1]);
            }
            if (c == '+' || c == '^' || c == '!') {
                if ((c == '+' || c == '^') && properties[properties.length - 1].length() > 1) {
                    throw new ParseError("Incorrect lexeme " + curLexeme.getValue() + " at position "
                        + (marker - curLexeme.getValue().length()));
                }
                readMethod();
                for (int i = 0; i < properties.length - 1; i++) {
                    Expression d = (Expression) pop(expStack);
                    getValue(d);
                }
                return;
            }
            propName = properties[properties.length - 1];
        }
        readCurLexeme();
        String defRef = readDefRef();
        MutableExpression e = new MutableExpression(nameConverter);
        e.setProperty(propName);
        e.setTarget(eval());
        // there is just one expression
        if (curLexeme != null && curLexeme.getType() == Lexeme.EQUALS_SIGN) {
            readCurLexeme();
            if (curLexeme == null) {
                throw new ParseError("At the end of the file there must be an expression following sign =");
            }
            Object sExp = readExpression();
            e.addArg(sExp);
            //there are some statements in braces or there is an indexed property
        } else {
            e = getExpressionWithStatements(e);
        }
        if (defRef != null) {
            environment.put(defRef, e);
        }
        getValue(e);
        //if there are several properties via dot (getters and setter)
        if (properties != null) {
            for (int i = 0; i < properties.length - 1; i++) {
                Expression d = (Expression) pop(expStack);
                getValue(d);
            }
        }
    }

    private void readIndexedGetterOrSetter() throws ParseError {
        int index;
        try {
            index = Integer.parseInt(curLexeme.getValue());
        } catch (NumberFormatException e) {
            throw new ParseError("Incorrect format of integer " + curLexeme.getValue());
        }
        readCurLexeme();
        MutableExpression e = new MutableExpression(nameConverter);
        e.setProperty(index);
        e.setTarget(eval());
        e.addArg(index);
        //indexed setter
        if (curLexeme.getType() == Lexeme.EQUALS_SIGN) {
            readCurLexeme();
            if (curLexeme == null) {
                throw new ParseError("At the end of the file there must be an expression following sign =");
            }
            Object sExp = readExpression();
            e.addArg(sExp);
            getValue(e);
            return;
        }
        //indexed getter
        if (curLexeme.getType() == Lexeme.COLON) {
            readCurLexeme();
            if (curLexeme == null) {
                throw new ParseError("At the end of the file there must be an object reference following sign :");
            }
            String defRef = readDefRef();
            //reading of the statement list
            e = getExpressionWithStatements(e);
            if (defRef != null) {
                environment.put(defRef, e);
            }
            getValue(e);
            return;
        }
        throw new ParseError("The lexeme at position " + (marker - curLexeme.getValue().length())
            + " must be = or : rather than " + curLexeme.getValue());
    }

    private Object readNewInstance() throws ParseError {
        readCurLexeme();
        if (curLexeme == null) {
            throw new ParseError("There is no object or array name following sign #");
        }
        if (curLexeme.getType() != Lexeme.NAME && curLexeme.getType() != Lexeme.ALIAS_SIGN) {
            throw new ParseError("The lexeme " + curLexeme.getValue() + " at position "
                + (marker - curLexeme.getValue().length()) + " must be the name of the object class");
        } else {
            String className = curLexeme.getValue();
            if (curLexeme.getType() == Lexeme.ALIAS_SIGN) {
                className = readAlias();
            }
            String arrayDimension = null;
            readCurLexeme();
            if (curLexeme != null && curLexeme.getType() == Lexeme.ARRAY_DIMENSION) {
                arrayDimension = curLexeme.getValue();
                readCurLexeme();
            }
            String defRef = readDefRef();
            MutableExpression e = new MutableExpression(nameConverter);
            e.setMethodName("new");
            //array
            if (arrayDimension != null) {
                if (curLexeme != null && curLexeme.getType() == Lexeme.DIMENSION_IN_ARRAY_TYPE) {
                    int dim = Integer.parseInt(curLexeme.getValue());
                    readCurLexeme();
                    className = processMultidimensionalArrayType(className, dim);
                }
                e.setTarget(Array.class);
                e.addArg(classForName2(nameConverter.demapClassName(className)));
                try {
                    e.addArg(Integer.parseInt(arrayDimension));
                } catch (NumberFormatException e1) {
                    throw new ParseError("Incorrect format of integer dimension " + arrayDimension);
                }
                //object
            } else {
                //reading of the argument list of object's constractor
                e.setTarget(classForName2(nameConverter.demapClassName(className)));
                ArrayList<Object> args = argList();
                if (args != null) {
                    for (Object arg : args) {
                        e.addArg(arg);
                    }
                }
            }
            //reading of the statement list of object or the element list of array
            e = getExpressionWithStatements(e);
            if (defRef != null) {
                environment.put(defRef, e);
            }
            return getValue(e);
        }
    }

    private String readAlias() throws ParseError {
        String packageName;
        String str = readIntAndDot();
        try {
            int number = Integer.parseInt(str);
            if (number >= aliases.size()) {
                throw new ParseError("Alias definition with number " + number + " does not exist");
            }
            packageName = aliases.get(number);
        } catch (NumberFormatException e) {
            throw new ParseError("Incorrect format of integer " + str);
        }
        // readName() is used because the lexeme Name can be mixed up with lexeme Boolean
        readName();
        if (curLexeme == null) {
            throw new ParseError("There is no object or array name following number");
        }
        if (curLexeme.getType() != Lexeme.NAME) {
            throw new ParseError("The lexeme " + curLexeme.getValue() + " at position "
                + (marker - curLexeme.getValue().length()) + " must be the name of the object class");
        }
        return packageName + "." + curLexeme.getValue();
    }

    // is used because the lexeme Int and sign Dot can be mixed up with the lexeme Double or Float
    private String readIntAndDot() throws ParseError {
        int i = 0;
        String num = "";
        while (i < lengthOfRest() && (Character.isDigit(nextSymOffset(i, 1).charAt(0)))) {
            num = num + nextSymOffset(i, 1);
            i++;
        }
        if (i >= lengthOfRest()) {
            throw new ParseError(
                "Incorrect format of alias: the end of the file is reached, but there is no sign . or number");
        }
        marker = marker + i;
        if (isCurSym('.')) {
            marker++;
        } else {
            throw new ParseError("The sign " + curSym(1) + " at position " + marker + " must be a dot");
        }
        if (lengthOfRest() == 0) {
            throw new ParseError(
                "Incorrect format of alias: the end of the file is reached, but there is no object name");
        }
        return num;
    }

    private String processMultidimensionalArrayType(String className, int dim) {
        String name = "";
        for (int i = 0; i < dim; i++) {
            name = name + "[";
        }
        if (className.equals("boolean")) {
            name = name + "Z";
        } else if (className.equals("byte")) {
            name = name + "B";
        } else if (className.equals("char")) {
            name = name + "C";
        } else if (className.equals("double")) {
            name = name + "D";
        } else if (className.equals("float")) {
            name = name + "F";
        } else if (className.equals("int")) {
            name = name + "I";
        } else if (className.equals("long")) {
            name = name + "J";
        } else if (className.equals("short")) {
            name = name + "S";
        } else {
            name = name + "L" + className + ";";
        }
        return name;
    }

    private String readDefRef() throws ParseError {
        String defRef = null;
        if (curLexeme != null && curLexeme.getType() == Lexeme.COLON) {
            readCurLexeme();
            if (curLexeme == null) {
                throw new ParseError("There is no object reference name following sign :");
            }
            if (curLexeme.getType() != Lexeme.NAME) {
                throw new ParseError("The lexeme " + curLexeme.getValue() + " at position "
                    + (marker - curLexeme.getValue().length()) + " must be the name of the reference to an object");
            }
            defRef = curLexeme.getValue();
            readCurLexeme();
        }
        return defRef;
    }

    private ArrayList<Object> argList() throws ParseError {
        ArrayList<Object> args = null;
        if (curLexeme != null && curLexeme.getType() == Lexeme.LEFT_PARENTHESIS) {
            readCurLexeme();
            if (curLexeme == null) {
                throw new ParseError("There are no expressions following sign (");
            }
            args = readExpressions();
            if (curLexeme == null) {
                throw new ParseError("The last lexeme must be )");
            }
            if (curLexeme.getType() != Lexeme.RIGHT_PARENTHESIS) {
                throw new ParseError("The lexeme " + curLexeme.getValue() + " at position "
                    + (marker - curLexeme.getValue().length()) + " must be )");
            }
            readCurLexeme();
        }
        return args;
    }

    private MutableExpression getExpressionWithStatements(MutableExpression e) throws ParseError {
        MutableExpression d = e;
        if (curLexeme != null && curLexeme.getType() == Lexeme.LEFT_BRACE) {
            expStack.add(d);
            readCurLexeme();
            if (curLexeme == null) {
                throw new ParseError("There are no statements following sign {");
            }
            readStatements();
            if (curLexeme == null) {
                throw new ParseError("The last lexeme must be }");
            }
            if (curLexeme.getType() != Lexeme.RIGHT_BRACE) {
                throw new ParseError("The lexeme " + curLexeme.getValue() + " at position "
                    + (marker - curLexeme.getValue().length()) + " must be }");
            }
            readCurLexeme();
            d = (MutableExpression) pop(expStack);
        }
        return d;
    }

    public Object getObject() {
        if (objects == null) {
            return null;
        } else {
            return objects.get(itemsRead++);
        }
    }

    private Object getValue(Expression exp) {
        try {
            return is.getValueResolver().getValue(exp);
        } catch (Exception e) {
            is.getExceptionListener().exceptionThrown(e);
            return null;
        }
    }

    private Object pop(ArrayList<MutableExpression> v) {
        int last = v.size() - 1;
        Object result = v.get(last);
        v.remove(last);
        return result;
    }

    private Object eval() {
        if (expStack.isEmpty()) { // this is about owner
            return is;
        }
        return getValue(expStack.get(expStack.size() - 1));
    }

    private Object getFieldValue(Object target, String fieldName) {
        try {
            Class<? extends Object> type = target.getClass();
            if (type == Class.class) {
                type = (Class<? extends Object>) target;
            }
            Field f = type.getField(fieldName);
            return f.get(target);
        } catch (Exception e) {
            is.getExceptionListener().exceptionThrown(e);
            return null;
        }
    }

    private Object classForName2(String name) {
        try {
            return is.getClassResolver().classForName(name, ldr);
        } catch (ClassNotFoundException e) {
            is.getExceptionListener().exceptionThrown(e);
        }
        return null;
    }

    public NameConverter getNameConverter() {
        return nameConverter;
    }

    public void setNameConverter(NameConverter nameConverter) {
        this.nameConverter = nameConverter;
    }
}

class ParseError extends Exception {
    // Represents a syntax error found in the user's input.
    ParseError(String message) {
        super(message);
    }
}

class MutableExpression extends MExpression {
    private Object target;
    private String methodName;

    private ArrayList<Object> argV = new ArrayList<Object>();
    private NameConverter nameCon;


    public MutableExpression(NameConverter n) {
        nameCon = n;
    }

    public Object[] getArguments() {
        return argV.toArray();
    }

    public String getMethodName() {
        if (property == null) {
            if (target instanceof Class) {
                return nameCon.demapMethodName(((Class) target).getName(), methodName, null);
            } else {
                if (target == null) {
                    throw new NullPointerException("target");
                }
                if (nameCon == null) {
                    throw new NullPointerException("nameCon");
                }
                return nameCon.demapMethodName(target.getClass().getName(), methodName, target);
            }
        }
        int setterArgs = (property instanceof String) ? 1 : 2;
        String methodName = (argV.size() == setterArgs) ? "set" : "get";
        if (property instanceof String) {
            return nameCon.demapMethodName(target.getClass().getName(), methodName + capitalize((String) property),
                target);
        } else {
            return methodName;
        }
    }

    public void addArg(Object arg) {
        argV.add(arg);
    }

    public void setTarget(Object target) {
        this.target = target;
    }

    public Object getTarget() {
        return target;
    }

    public void setMethodName(String methodName) {
        this.methodName = methodName;
    }

    public void setValue(Object value) {
        super.setValue(value);
    }

    public Object getValue() throws Exception {
        return super.getValue();
    }

    private String capitalize(String propertyName) {
        if (propertyName.length() == 0) {
            return propertyName;
        }
        return propertyName.substring(0, 1).toUpperCase() + propertyName.substring(1);
    }

}
