/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this file,
 * You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Portions created by Devexperts are Copyright (C) 2007-2012 Devexperts.
 */

package com.devexperts.evolution.utils.configencoder;

import java.util.ArrayList;

/**
 * Class DefaultNameConverter maps and demaps class, package, method names. Rules are:<br/>
 * <br/>
 * 1. At the beginning class and package names are mapped in order how they go in method
 * addClassAndPackageMapping()<br/>
 * 2. Then method names are mapped.<br/>
 * <br/>
 * It means that at encoding a method name is changed in context of the object which name has not been encoded and at
 * decoding a method name is changed in context of the object which name has been decoded. if you demap just class or
 * package name then you can change class name, but if you map method name then you can change just package name, class
 * name must be left the same.
 *
 * @author Anton V. Shakhmin
 * @see MappingRule
 */
public class DefaultNameConverter implements NameConverter {
    protected ArrayList<MappingRule> classRuleList = new ArrayList<MappingRule>();
    protected ArrayList<MappingRule> methodRuleList = new ArrayList<MappingRule>();

    public void addMethodRule(String oldName, String newName, String oldClassName) {
        addMethodRule(new MappingRule(oldName, newName, oldClassName));
    }

    public void addMethodRule(MappingRule o) {
        methodRuleList.add(o);
    }

    public void addMethodRule(String oldName, String newName, Class oldClassName) {
        addMethodRule(oldName, newName, oldClassName.getName());
    }

    public void addClassRule(String oldName, String newName) {
        classRuleList.add(new MappingRule(oldName, newName, "class"));
    }

    public void addClassRule(Class oldName, String newName) {
        addClassRule(oldName.getName(), newName);
    }

    public void addPackageRule(String oldPackageName, String newPackageName) {
        classRuleList.add(new MappingRule(oldPackageName, newPackageName, "package"));
    }

    public String mapClassName(String oldName) {
        String res = oldName;
        for (MappingRule aRuleList : classRuleList) {
            if (aRuleList.getClassName().equals("class")) {
                if (oldName.equals(aRuleList.getOldName())) {
                    res = aRuleList.getNewName();
                    break;
                }
            } else {
                if (extractPackageName(oldName).equals(aRuleList.getOldName())) {
                    res = oldName.replaceFirst(aRuleList.getOldName(), aRuleList.getNewName());
                    break;
                }
            }
        }
        return res;
    }

    public String mapMethodName(String className, String oldName) {
        String res = oldName;
        for (MappingRule aRuleList : methodRuleList) {
            if (aRuleList.getOldName().equals(oldName) && aRuleList.getClassName().equals(className)) {
                res = aRuleList.getNewName();
                break;
            }
        }
        return res;
    }

    public String demapClassName(String newName) {
        String res = newName;
        for (MappingRule aRuleList : classRuleList) {
            if (aRuleList.getClassName().equals("class")) {
                if (newName.equals(aRuleList.getNewName())) {
                    res = aRuleList.getOldName();
                    break;
                }
            } else {
                if (extractPackageName(newName).equals(aRuleList.getNewName())) {
                    res = newName.replaceFirst(aRuleList.getNewName(), aRuleList.getOldName());
                    break;
                }
            }
        }
        return res;
    }

    public String demapMethodName(String className, String newName, Object target) {
        String res = newName;
        for (MappingRule aRuleList : methodRuleList) {
            if (aRuleList.getNewName().equals(newName)
                && aRuleList.demapMethodMatchClass(className, target))
            {
                res = aRuleList.getOldName();
                break;
            }
        }
        return res;
    }

    private static String extractPackageName(String name) {
        if (name.lastIndexOf('.') == -1) {
            return "";
        }
        return name.substring(0, name.lastIndexOf('.'));
    }

    public static String getClassSimpleName(String name) {
        int index = name.lastIndexOf('.');
        if (index == -1) {
            return "";
        }
        return name.substring(index + 1);
    }
}
